from typing import Tuple

from fretboardgtr.constants import FLAT_CHROMATICS_NOTES, SHARP_CHROMATICS_NOTES


class Note:
    _BASE_NOTES = ["A", "B", "C", "D", "E", "F", "G"]
    _BASE_NOTES_DISTANCE = [2, 2, 1, 2, 2, 1, 2]

    def __init__(self, name: str):
        self.name = name
        if not self.check_if_valid():
            raise ValueError(f"{name} is not a valid note")

    def __str__(self) -> str:
        return self.name

    def base_note(self) -> str:
        return self.name[0]

    def _resolve(self, prefer_flat: bool = True) -> str:
        if len(self.name) == 1:
            return self.name

        chromatic_scale = (
            FLAT_CHROMATICS_NOTES if prefer_flat else SHARP_CHROMATICS_NOTES
        )
        if self.name in chromatic_scale:
            return self.name

        base_note, alterations = self.name[0:1], self.name[1:]
        new_note = base_note
        if "b" in alterations:
            new_note = chromatic_scale[
                (chromatic_scale.index(base_note) - (len(alterations))) % 12
            ]
        elif "#" in alterations:
            new_note = chromatic_scale[
                (chromatic_scale.index(base_note) + (len(alterations))) % 12
            ]

        return new_note

    def resolve(self, prefer_flat: bool = True) -> "Note":
        """Resolve alterations in notes."""
        return Note(self._resolve(prefer_flat))

    def check_if_valid(self) -> bool:
        resolved_note = self._resolve()
        return (
            resolved_note in SHARP_CHROMATICS_NOTES
            or resolved_note in FLAT_CHROMATICS_NOTES
        )

    def sharpen(self) -> "Note":
        """Sharpen the note name by increasing the pitch by one semitone.

        Returns
        -------
        Note
            The sharpened note
        """
        # If the note name has only one letter, just add a "#" at the end
        if len(self.name) == 1:
            return Note(self.name + "#")

        base_note, alterations = self.name[0:1], self.name[1:]

        # If the note has a flat alteration, remove it
        if "b" in alterations:
            return Note(base_note + alterations[:-1])

        if "#" in alterations:
            return Note(self.name + "#")

        return self

    def flatten(self) -> "Note":
        """Flatten the note name by decreasing the pitch by one semitone.

        Returns
        -------
        Note
            The flattened note
        """
        # If the note has only one character, just add 'b' to it
        if len(self.name) == 1:
            return Note(self.name + "b")

        base_note, alterations = self.name[0:1], self.name[1:]

        if "b" in alterations:
            return Note(self.name + "b")

        # If the note has '#' in its alterations, remove the sharp
        if "#" in alterations:
            return Note(base_note + alterations[:-1])

        return self

    def __next_base_note(self, base_note: str) -> Tuple[str, int]:
        idx_base_note = self._BASE_NOTES.index(base_note)
        next_idx = (idx_base_note + 1) % 7
        return self._BASE_NOTES[next_idx], self._BASE_NOTES_DISTANCE[next_idx]

    def __previous_base_note(self, base_note: str) -> Tuple[str, int]:
        idx_base_note = self._BASE_NOTES.index(base_note)
        next_idx = (idx_base_note - 1) % 7
        # We add 1 to the note distance because as we're going backward,
        # we need to use the n+1 interval
        return self._BASE_NOTES[next_idx], self._BASE_NOTES_DISTANCE[(next_idx + 1) % 7]

    def flat_enharmonic(self) -> "Note":
        """Transform the note into its enharmonic flat equivalent.

        If the note has alterations, it retains them and applies flats
        instead of sharps.

        Returns
        -------
        Note
            The enharmonic equivalent note with flats.
        """
        # Get the base note and any alterations
        base_note = self.name[0:1]
        alterations = self.name[1:] if len(self.name) >= 2 else None

        # Apply alterations
        if alterations is not None and "#" in alterations:
            return self.resolve(prefer_flat=True)

        target_note, distance = self.__next_base_note(base_note)

        if alterations is not None and "b" in alterations:
            return Note(target_note + alterations + "b" * distance)
        else:
            flats_to_add = "b" * distance
            return Note(target_note + flats_to_add)

    def sharp_enharmonic(self) -> "Note":
        """Transform the note into its enharmonic sharp equivalent.

        If the note has alterations, it retains them and applies sharps
        instead of flats.

        Returns
        -------
        Note
            The enharmonic equivalent note with sharps.
        """
        # Get the base note and any alterations
        base_note = self.name[0:1]
        alterations = self.name[1:] if len(self.name) >= 2 else None

        # Apply alterations
        if alterations is not None and "b" in alterations:
            return self.resolve(prefer_flat=False)

        target_note, distance = self.__previous_base_note(base_note)

        if alterations is not None and "#" in alterations:
            return Note(target_note + alterations + "#" * distance)
        else:
            sharps_to_add = "#" * distance
            return Note(target_note + sharps_to_add)



import pytest

from fretboardgtr.notes import Note


@pytest.mark.parametrize("invalid_note", ["Z", "H", "Invalid"])
def test_invalid_note_creation(invalid_note):
    with pytest.raises(ValueError):
        Note(invalid_note)


@pytest.mark.parametrize(
    "valid_note",
    ["C", "C#", "C###" "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"],
)
def test_valid_note_creation(valid_note):
    note = Note(valid_note)
    assert str(note) == valid_note


@pytest.mark.parametrize(
    "note_str, expected_resolved_note, prefer_flat",
    [
        ("Fb", "E", False),
        ("D#", "D#", False),
        ("Ab", "Ab", True),
        ("Abbb", "Gb", True),
        ("F#", "F#", False),
        ("E#", "F", False),
        ("G#", "G#", False),
        ("A#", "A#", False),
        ("B#", "C", False),
    ],
)
def test_resolve(note_str, expected_resolved_note, prefer_flat):
    note = Note(note_str)
    resolved_note = note.resolve(prefer_flat)
    assert str(resolved_note) == expected_resolved_note


@pytest.mark.parametrize(
    "note_str, expected_sharpened_note",
    [
        ("D", "D#"),
        ("C", "C#"),
        ("Db", "D"),
        ("Dbb", "Db"),
        ("Dbbbb", "Dbbb"),
        ("D#", "D##"),
    ],
)
def test_sharpen(note_str, expected_sharpened_note):
    note = Note(note_str)
    sharpened_note = note.sharpen()
    assert str(sharpened_note) == expected_sharpened_note


@pytest.mark.parametrize(
    "note_str, expected_flattened_note",
    [
        ("D", "Db"),
        ("C", "Cb"),
        ("Db", "Dbb"),
        ("D#", "D"),
    ],
)
def test_flatten(note_str, expected_flattened_note):
    note = Note(note_str)
    flattened_note = note.flatten()
    assert str(flattened_note) == expected_flattened_note


@pytest.mark.parametrize(
    "note_str, expected_flat_enharmonic_note",
    [
        ("A#", "Bb"),
        ("A##", "B"),
        ("A###", "C"),
        ("A####", "Db"),
        ("A", "Bbb"),
        ("Ab", "Bbbb"),
        ("G#", "Ab"),
        ("G", "Abb"),
        ("Gb", "Abbb"),
        ("F#", "Gb"),
        ("F", "Gbb"),
        ("E", "Fb"),
        ("E#", "F"),
        ("Eb", "Fbb"),
    ],
)
def test_flat_enharmonic(note_str, expected_flat_enharmonic_note):
    note = Note(note_str)
    flat_enharmonic_note = note.flat_enharmonic()
    assert str(flat_enharmonic_note) == expected_flat_enharmonic_note


@pytest.mark.parametrize(
    "note_str, expected_sharp_enharmonic_note",
    [
        ("Ab", "G#"),
        ("Abb", "G"),
        ("Abbb", "F#"),
        ("Abbbb", "F"),
        ("A", "G##"),
        ("A#", "G###"),
        ("Gb", "F#"),
        ("G", "F##"),
        ("G#", "F###"),
        ("Fb", "E"),
        ("F", "E#"),
        ("E", "D##"),
        ("Eb", "D#"),
    ],
)
def test_sharp_enharmonic(note_str, expected_sharp_enharmonic_note):
    note = Note(note_str)
    sharp_enharmonic_note = note.sharp_enharmonic()
    assert str(sharp_enharmonic_note) == expected_sharp_enharmonic_note



from dataclasses import dataclass
from itertools import product
from typing import List, Optional

from fretboardgtr.constants import CHORDS_DICT_ESSENTIAL, CHROMATICS_NOTES, SCALES_DICT
from fretboardgtr.utils import chromatic_position_from_root, get_note_from_index


def find_first_index(_list: List[int], value: int) -> Optional[int]:
    try:
        index = _list.index(value)
        return index
    except ValueError:
        return None  # If the value is not found in the tuple, return None


@dataclass
class NotesContainer:
    root: str
    notes: List[str]

    def get_scale(self, tuning: List[str], max_spacing: int = 5) -> List[List[int]]:
        """Get the scale of each string in the given tuning.

        Goes from 0 up to (12 + max_spacing - 1) on the fretboard

        Parameters
        ----------
        tuning (List[str])
            The tuning of each string.

        Returns
        -------
        List[List[int]]
            The scale of each string in the tuning.
        """
        scale = []

        # Iterate over each string in the tuning
        for string_note in tuning:
            indices = []

            # Iterate over each note in the self.notes list
            for note in self.notes:
                _idx = chromatic_position_from_root(note, string_note)

                # Add the chromatic positions of the note on the string
                # until it reaches or exceeds the maximum position of 16
                while _idx <= 12 + max_spacing - 1:
                    indices.append(_idx)
                    _idx += 12

            # Sort the indices in ascending order
            scale.append(sorted(indices))

        return scale

    def get_chord_fingerings(
        self,
        tuning: List[str],
        max_spacing: int = 5,
        min_notes_in_chord: int = 2,
        number_of_fingers: int = 4,
    ) -> List[List[Optional[int]]]:
        """Get all probably possible fingering for a specific tuning.

        Parameters
        ----------
        tuning : List[str]
            List of note of the tuning
        max_spacing : int
            Maximum spacing between notes
        min_notes_in_chord : int
            Minimum number of notes in chord
        number_of_fingers : int
            Number of fingers allowed

        Returns
        -------
        List[List[Optional[int]]]
            List of propably possible fingerings
        """
        scale = self.get_scale(tuning, max_spacing)

        fingerings = []
        for combination in product(*scale):
            non_zero_numbers = [num for num in combination if num != 0]
            # No more than 4 fingers but duplicated allowed
            if len(set(non_zero_numbers)) > number_of_fingers:
                continue

            new_combination = list(combination)
            while True:
                # If 0 note or only one this is not a chord so break
                if len(non_zero_numbers) < min_notes_in_chord:
                    break
                # If the spacing is less than 5 then it'ok so break
                if max(non_zero_numbers) - min(non_zero_numbers) <= max_spacing:
                    break

                # If the spacing is more than 5 then remplace min values by None
                index_of_min = find_first_index(new_combination, min(non_zero_numbers))
                index_of_non_zero_min = find_first_index(
                    non_zero_numbers, min(non_zero_numbers)
                )
                if index_of_min is not None and index_of_non_zero_min is not None:
                    new_combination[index_of_min] = None
                    del non_zero_numbers[index_of_non_zero_min]

            notes = []
            for index, note in zip(new_combination, tuning):
                if index is not None:
                    notes.append(get_note_from_index(index, note))

            # Each notes should appear at least once.
            if set(notes) != set(self.notes):
                continue

            fingerings.append(list(new_combination))
        return fingerings

    def get_scale_positions(
        self,
        tuning: List[str],
        max_spacing: int = 5,
    ) -> List[List[List[Optional[int]]]]:
        """Get all possible scale positions for a specific tuning.

        Parameters
        ----------
        tuning : List[str]
            List of note of the tuning
        max_spacing : int
            Maximum spacing between notes

        Returns
        -------
        List[List[List[Optional[int]]]]
            List of all possible scale positions
        """
        scale = self.get_scale(tuning, max_spacing)
        fingerings: List[List[List[Optional[int]]]] = []
        for first_string_pos in scale[0]:
            fingering: List[List[Optional[int]]] = []
            for string in scale:
                string_fingering: List[Optional[int]] = []
                for note in string:
                    if (
                        note - first_string_pos < 0
                        or note - first_string_pos >= max_spacing
                    ):
                        continue
                    string_fingering.append(note)
                fingering.append(string_fingering)
            fingerings.append(fingering)
        return fingerings


class ScaleFromName:
    """Object that generating NotesContainer object from root and mode.

    Given a root name and a mode name, get the resulting scale.

    Also :
    Mode name can be given thanks to the constants.ModeName enum as well as string
    Note name can be given thanks to the constants.NoteName enum as well as string

    Example
    -------
    >>> ScaleFromName(root='C',mode='Dorian').build()
        NotesContainer(root= 'C', scale = ['C', 'D', 'D#', 'F', 'G', 'A', 'A#'])
    >>> ScaleFromName(root=Name.C,mode=ModeName.DORIAN).build()
        NotesContainer(root= 'C', scale = ['C', 'D', 'D#', 'F', 'G', 'A', 'A#'])
    """

    def __init__(self, root: str = "C", mode: str = "Ionian"):
        self.root = root
        self.mode = mode

    def build(self) -> NotesContainer:
        index = CHROMATICS_NOTES.index(self.root)
        mode_idx = SCALES_DICT[self.mode]
        scale = []
        for note_id in mode_idx:
            scale.append(CHROMATICS_NOTES[(index + note_id) % 12])
        return NotesContainer(self.root, scale)


class ChordFromName:
    """Object generating NotesContainer object from root and chord quality.

    Given a root name and a quality name, get the resulting scale.

    Also :
    Mode name can be given thanks to the constants.ChordName enum as well as string
    Note name can be given thanks to the constants.NoteName enum as well as string

    Example
    -------
    >>> ChordFromName(root='C',quality='M').build()
        NotesContainer(root= 'C', scale = ['C', 'E', 'G'])
    >>> ChordFromName(root=NoteName.C,quality=ChordName.MAJOR).build()
        NotesContainer(root= 'C', scale = ['C', 'E', 'G'])
    """

    def __init__(self, root: str = "C", quality: str = "M"):
        self.root = root
        self.quality = quality

    def build(self) -> NotesContainer:
        index = CHROMATICS_NOTES.index(self.root)

        quality_idx = CHORDS_DICT_ESSENTIAL[self.quality]
        scale = []
        for note_id in quality_idx:
            scale.append(CHROMATICS_NOTES[(index + note_id) % 12])
        return NotesContainer(self.root, scale)



from fretboardgtr.notes_creators import ChordFromName, ScaleFromName


def test_scale_creator():
    scale = ScaleFromName(root="C", mode="Ionian").build()
    assert scale.notes == ["C", "D", "E", "F", "G", "A", "B"]


def test_chord_creator():
    chord = ChordFromName(root="C", quality="M").build()
    assert chord.notes == ["C", "E", "G"]


def test_chord_creator_fingerings():
    fingerings = (
        ChordFromName(root="C", quality="M")
        .build()
        .get_chord_fingerings(["E", "A", "D", "G", "B", "E"])
    )
    assert len(fingerings) > 1000


def test_scale_creator_position():
    scale_positions = (
        ScaleFromName(root="C", mode="Ionian")
        .build()
        .get_scale_positions(["E", "A", "D", "G", "B", "E"])
    )
    assert len(scale_positions) > 5
